#include "CHCPlusPlusTraverser.h"
#include "RenderState.h"

using namespace std;

namespace CHCDemo
{


CHCPlusPlusTraverser::CHCPlusPlusTraverser()
{
}


void CHCPlusPlusTraverser::Traverse()
{
	//-- PART 1: process finished occlusion queries
	while (!mDistanceQueue.empty() || !mQueryQueue.empty())
	{
		bool resultAvailable;
		while (!mQueryQueue.empty() && 
			   (mDistanceQueue.empty() || (resultAvailable = mQueryQueue.front()->ResultAvailable())))
		{
			while (!resultAvailable && !mVQueue.empty())
			{
				BvhNode *node = mVQueue.front();
				mVQueue.pop();
				
				OcclusionQuery *query = IssueOcclusionQuery(node, true);
				mQueryQueue.push(query);
				resultAvailable = mQueryQueue.front()->ResultAvailable();
			}
			
			OcclusionQuery *query = mQueryQueue.front();
			mQueryQueue.pop();

			// wait until result available
			int visiblePixels = query->GetQueryResult();

			if (visiblePixels > mVisibilityThreshold)
			{
				BvhNode *node = query->GetFrontNode();

				node->SetVisible(true);
				mBvh->MakeParentsVisible(node);
				TraverseNode(node);
			}
			else
			{
				++ mStats.mNumQueryCulledNodes;
			}
		}	

		//-- PART 2: hierarchical traversal
		if (!mDistanceQueue.empty())
		{
			BvhNode *node = mDistanceQueue.top();
			mDistanceQueue.pop();
	
			if (mBvh->IsWithinViewFrustum(node))
			{
				// for near plane intersecting bounding box possible 
				// wrong results => skip occlusion query
				if (IntersectsNearPlane(node))
				{
					// update node's visited flag
					node->SetLastVisitedFrame(mFrameId);
					node->SetVisible(true);
					mBvh->MakeParentsVisible(node);

					TraverseNode(node);
				}
				else
				{		
					// identify previously visible nodes
					const bool wasVisible = node->IsVisible() && (node->GetLastVisitedFrame() == mFrameId - 1);
					
					// identify nodes that we cannot skip queries for
					const bool testFeasible = (!wasVisible || (node->IsVirtualLeaf() && (node->GetAssumedVisibleFrameId() <= mFrameId)));
						 

					// update node's visited flag
					node->SetLastVisitedFrame(mFrameId);
					
					// skip testing previously visible interior nodes
					if (testFeasible)
					{
						// reset node's visibility classification 
						node->SetVisible(false);

						if (!wasVisible)
						{
							QueryPreviouslyInvisibleNodes(node);
						}
						else
						{
							mVQueue.push(node);
						}
					}
					
					// always traverse a node if it was previously visible
					if (wasVisible)
						TraverseNode(node);
				}
			}
			else
			{
				// for stats
				++ mStats.mNumFrustumCulledNodes;
			}
		}

		// distance queue empty: feed the remaining multiqueries to
		// be able to proceed traversal
		if (mDistanceQueue.empty())
		{
			IssueMultiQueries();
		}
	}

	// render the rest of the objects
	if (mUseRenderQueue)
	{
		if (mRenderState->SetState(RenderState::RENDER))
			++ mStats.mNumStateChanges;

		mRenderQueue.Render();
		mRenderQueue.Clear();
	}

	//////////////
	//-- issue remaining previously visible node queries

	while (!mVQueue.empty())
	{
		BvhNode *node = mVQueue.front();
		mVQueue.pop();

		OcclusionQuery *query = IssueOcclusionQuery(node, true);
		mQueryQueue.push(query);
	}

	while (!mQueryQueue.empty())
	{
		OcclusionQuery *query = mQueryQueue.front();
		mQueryQueue.pop();

		// wait until result available
		int visiblePixels = query->GetQueryResult();

		if (visiblePixels > mVisibilityThreshold)
		{
			BvhNode *node = query->GetFrontNode();

			node->SetVisible(true);
			mBvh->MakeParentsVisible(node);
		}
		else
		{
			++ mStats.mNumQueryCulledNodes;
		}
	}
}



void CHCPlusPlusTraverser::QueryPreviouslyInvisibleNodes(BvhNode *node)
{
	mIQueue.push(node);

	if (mIQueue.size() > mMaxBatchSize)
	{
		IssueMultiQueries();
	}
}


void CHCPlusPlusTraverser::IssueMultiQueries()
{
	if (0 && mUseRenderQueue)
	{
		if (mRenderState->SetState(RenderState::RENDER))
			++ mStats.mNumStateChanges;
	
		mRenderQueue.Render();
		mRenderQueue.Clear();
	}

	while (!mIQueue.empty())
	{
		BvhNode *node = mIQueue.front();
		mIQueue.pop();

		OcclusionQuery *query = IssueOcclusionQuery(node, false);
		mQueryQueue.push(query);
	}
}


}